#include <glad/glad.h>
#include <SDL.h>
#include <toolbox.hpp>
#include <wave_handler.hpp>
#include <infrastructure.hpp>
#include <iostream>
#include <input_handler.hpp>
#include <gui_handler.hpp>
#include <slim_blockchain_handler.hpp>
#include <drawing_handler.hpp>
#include <chrono>
#include <thread>
#include <pattern_handler.hpp>
#include <timeout_handler.hpp>

/*Screen resolution*/
#define WIDTH 2560
#define HEIGHT 1440
//#define WIDTH 3840
//#define HEIGHT 2160

/*Path to resources*/

int main(int argc, char** argv) {
	std::string top_path = "C:\\Users\\engel\\VS_Projects\\FocusTerra\\framebuffer-testing\\";
	std::string rsc_path = top_path + "resource\\";
	std::string img_path = rsc_path + "images\\";
	std::string tex_path = rsc_path + "textures\\home\\";
	std::string font_path = top_path + "include\\imgui\\misc\\fonts\\";
	std::string shd_path = top_path + "shaders\\";
	/*
	int tex_offscreen_left = 0;
	int tex_offscreen_right = 450;
	int tex_offscreen_bottom = 450;
	int tex_offscreen_top = 450;
	*/
	int tex_offscreen_left = 0;
	int tex_offscreen_right = 600;
	int tex_offscreen_bottom = 600;
	int tex_offscreen_top = 600;
	int texwidth = WIDTH + tex_offscreen_left + tex_offscreen_right;
	int texheight = HEIGHT + tex_offscreen_bottom + tex_offscreen_top;

	Toolbox tb(WIDTH, HEIGHT, texwidth, texheight, tex_offscreen_left, tex_offscreen_right, tex_offscreen_bottom, tex_offscreen_top, shd_path, tex_path, rsc_path);
	WaveHandler waves(tb);

	//waves.initialize(rsc_path + "damp_exp", rsc_path + "ft_palette");
	waves.initialize(tex_path + "bare", rsc_path + "ft_palette");
	waves.generate_and_transfer_textures(tb);
	InputHandler ioHandler;
	GuiHandler guiHandler(0.15f);
	guiHandler.init(tb, img_path, font_path);
	SlimBlockchainHandler bch(tb);
	DrawingHandler drah(tb);
	PatternHandler paha(tb);
	TimeoutHandler tiha(600);
	
	/*Scale the gui*/
	ImGui::GetStyle().ScaleAllSizes(2.f);

	while (tb.g_state != static_cast<int>(GSTATE::EXIT)) {
		/*Start New Frame*/
		//std::cout << "New Frame" << std::endl;
		tb.newFrame();
		//std::cout << "Error: " << glGetError() << std::endl;
		/*Input Handler*/
		//std::cout << "Handling Input" << std::endl;
		ioHandler.update(tb);
		//std::cout << "Error: " << glGetError() << std::endl;
		/*Timeout Handler*/
		tiha.update(tb);
		/*Gui Handler*/
		//std::cout << "Handling Gui" << std::endl;
		guiHandler.update(tb);
		//std::cout << "Error: " << glGetError() << std::endl;
		/*Waves Handler update*/
		waves.update(tb);
		/*Blockchain Handler*/
		//std::cout << "Handling Blockchain" << std::endl;
		bch.update(tb);
		//std::cout << "Error: " << glGetError() << std::endl;
		/*Drawing Manager*/
		drah.update(tb);
		/*Pattern Handler*/
		paha.update(tb);

		/*Waves Handler stepping*/
		//std::cout << "Stepping Wave" << std::endl;
		waves.prepare_step();
		waves.step(tb, 2);
		//std::cout << "Error: " << glGetError() << std::endl;

		/*Render*/
		//std::cout << "Rendering Wave" << std::endl;
		waves.render();
		//std::cout << "Error: " << glGetError() << std::endl;
		//std::cout << "Rendering Gui" << std::endl;
		guiHandler.render(tb);
		//std::cout << "Error: " << glGetError() << std::endl;
		//std::cout << "Swapping Window" << std::endl;
		SDL_GL_SwapWindow(tb.infra.window());
		//std::cout << "Error: " << glGetError() << std::endl;

		/*Debugging*/
		//int w_width, w_height;
		//SDL_GetWindowSize(tb.infra.window(), &w_width, &w_height);
		size_t num_ev = tb.events.size();
		size_t num_mess = tb.mailbox.size();
		
		/*
		if (num_ev != 0)
			std::cout << "Events: " << num_ev << std::endl;
		if (num_mess != 0)
			std::cout << "Messages: " << num_mess << std::endl;
		*/

		//std::cout << "Window Dimensions: " << w_width << " x " << w_height << std::endl;
		GLint err = glGetError();
		if (err != GL_NO_ERROR) {
			std::cout << "Error Code: " << err << std::endl;
			std::cout << "GL_NO_ERROR: " << GL_NO_ERROR << std::endl;
			std::cout << "GL_INVALID_ENUM: " << GL_INVALID_ENUM << std::endl;
			std::cout << "GL_INVALID_VALUE: " << GL_INVALID_VALUE << std::endl;
			std::cout << "GL_INVALID_OPERATION: " << GL_INVALID_OPERATION << std::endl;
			std::cout << "GL_INVALID_FRAMEBUFFER_OPERATION: " << GL_INVALID_FRAMEBUFFER_OPERATION << std::endl;
			std::cout << "GL_OUT_OF_MEMORY: " << GL_OUT_OF_MEMORY << std::endl;
		}

		//std::this_thread::sleep_for(std::chrono::milliseconds(500));
	}

}